@@ -65,6 +65,7 @@ struct _EggPlayPreviewPrivate {
 	GtkWidget *time_label;
 
 	GstElement *playbin;
+	GstQuery   *query;
 	GstState state;
 
 	gchar *play_icon_name;
@@ -73,11 +74,12 @@ struct _EggPlayPreviewPrivate {
 	gchar *artist;
 	gchar *album;
 
-	gint duration;
-	gint position;
+	gint64 duration;
+	gint64 position;
 
 	gint timeout_id;
 
+	gboolean seeking;
 	gboolean is_seekable;
 
 	gchar *uri;
@@ -107,16 +109,21 @@ static gboolean _change_value_cb             (GtkRange *range,
 											  EggPlayPreview *play_preview);
 static void _clicked_cb                      (GtkButton      *button,
 											  EggPlayPreview *play_preview);
+static void _duration_notify_cb              (EggPlayPreview *play_preview,
+											  GParamSpec     *pspec,
+											  gpointer        user_data);
+static void _style_updated_cb                (GtkWidget *widget,
+											  gpointer   user_data);
 static void _setup_pipeline                  (EggPlayPreview *play_preview);
 static void _clear_pipeline                  (EggPlayPreview *play_preview);
 static gboolean _process_bus_messages        (GstBus         *bus,
 											  GstMessage     *msg,
 											  EggPlayPreview *play_preview);
 static gboolean _query_seeking               (GstElement *element);
-static gint _query_duration                  (GstElement *element);
-static gint _query_position                  (GstElement *element);
-static void _seek                            (GstElement *element,
-											  gint position);
+static gint64 _query_duration                (GstElement *element);
+static void _seek                            (EggPlayPreview *play_preview,
+											  GstElement *element,
+											  gint64      position);
 static gboolean _is_playing                  (GstState state);
 static void _play                            (EggPlayPreview *play_preview);
 static void _pause                           (EggPlayPreview *play_preview);
@@ -207,23 +214,23 @@ egg_play_preview_class_init (EggPlayPreviewClass *klass)
 
 	g_object_class_install_property (gobject_class,
 									 PROP_POSITION,
-									 g_param_spec_int ("position",
-													   _("Position"),
-													   _("The position in the current stream in seconds."),
-													   0, G_MAXINT, 0,
-													   G_PARAM_READWRITE |
-													   G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
-													   G_PARAM_STATIC_BLURB));
+									 g_param_spec_int64 ("position",
+														 _("Position"),
+														 _("The position in the current stream in seconds."),
+														 0, G_MAXINT, 0,
+														 G_PARAM_READWRITE |
+														 G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
+														 G_PARAM_STATIC_BLURB));
 
 	g_object_class_install_property (gobject_class,
 									 PROP_DURATION,
-									 g_param_spec_int ("duration",
-													   _("Duration"),
-													   _("The duration of the current stream in seconds."),
-													   0, G_MAXINT, 0,
-													   G_PARAM_READABLE |
-													   G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
-													   G_PARAM_STATIC_BLURB));
+									 g_param_spec_int64 ("duration",
+														 _("Duration"),
+														 _("The duratio	n of the current stream in seconds."),
+														 0, G_MAXINT, 0,
+														 G_PARAM_READABLE |
+														 G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
+														 G_PARAM_STATIC_BLURB));
 
 	gst_init (NULL, NULL);
 }
@@ -234,9 +241,7 @@ egg_play_preview_init (EggPlayPreview *play_preview)
 	EggPlayPreviewPrivate *priv;
 	PangoAttribute *bold;
 	PangoAttrList *attrs;
-	GtkWidget *vbox;
-	GtkWidget *hbox;
-	GtkWidget *align;
+	GtkGrid *grid;
 
 	play_preview->priv = priv = GET_PRIVATE (play_preview);
 
@@ -253,15 +258,19 @@ egg_play_preview_init (EggPlayPreview *play_preview)
 
 	priv->uri = NULL;
 
-	gtk_box_set_homogeneous (GTK_BOX (play_preview), FALSE);
-	gtk_box_set_spacing (GTK_BOX (play_preview), 6);
+	grid = (GtkGrid*) gtk_grid_new ();
+	g_object_set (grid,
+				  "column-spacing", 12,
+				  "border-width", 12,
+				  NULL);
 
 	/* track info */
-	vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
-
 	priv->title_label = gtk_label_new (NULL);
-	gtk_label_set_justify (GTK_LABEL (priv->title_label), GTK_JUSTIFY_LEFT);
-	gtk_misc_set_alignment (GTK_MISC (priv->title_label), 0.0, 0.5);
+	g_object_set (priv->title_label,
+				  "justify", GTK_JUSTIFY_LEFT,
+				  "valign", GTK_ALIGN_CENTER,
+				  "xalign", 0.0,
+				  NULL);
 	bold = pango_attr_weight_new (PANGO_WEIGHT_BOLD);
 	attrs = pango_attr_list_new ();
 	bold->start_index = 0;
@@ -269,14 +278,15 @@ egg_play_preview_init (EggPlayPreview *play_preview)
 	pango_attr_list_insert (attrs, bold);
 	gtk_label_set_attributes (GTK_LABEL (priv->title_label), attrs);
 	pango_attr_list_unref(attrs);
-	gtk_box_pack_start (GTK_BOX (vbox), priv->title_label, TRUE, TRUE, 0);
+	gtk_grid_attach (grid, priv->title_label, 0, 0, 1, 1);
 
 	priv->artist_album_label = gtk_label_new (NULL);
-	gtk_label_set_justify (GTK_LABEL (priv->artist_album_label), GTK_JUSTIFY_LEFT);
-	gtk_misc_set_alignment (GTK_MISC (priv->artist_album_label), 0.0, 0.5);
-	gtk_box_pack_start (GTK_BOX (vbox), priv->artist_album_label, TRUE, TRUE, 0);
-
-	hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 12);
+	g_object_set (priv->artist_album_label,
+				  "justify", GTK_JUSTIFY_LEFT,
+				  "valign", GTK_ALIGN_CENTER,
+				  "xalign", 0.0,
+				  NULL);
+	gtk_grid_attach (grid, priv->artist_album_label, 0, 1, 1, 1);
 
 	/* play button */
 	priv->play_button = gtk_button_new ();
@@ -286,27 +296,39 @@ egg_play_preview_init (EggPlayPreview *play_preview)
 	    priv->play_icon_name = "media-playback-start";
 	priv->play_button_image = gtk_image_new_from_icon_name (priv->play_icon_name, GTK_ICON_SIZE_BUTTON);
 	gtk_container_add (GTK_CONTAINER (priv->play_button), priv->play_button_image);
-	align = gtk_alignment_new (0.5, 0.5, 1.0, 0.0);
-	gtk_container_add (GTK_CONTAINER (align), priv->play_button);
+	g_object_set (priv->play_button,
+				  "valign", GTK_ALIGN_CENTER,
+				  NULL);
+	gtk_grid_attach (grid, priv->play_button, 1, 0, 1, 2);
 
 	/* time scale */
 	priv->time_scale = gtk_scale_new_with_range(GTK_ORIENTATION_HORIZONTAL, 0.0, 100.0, 1.0);
 	gtk_scale_set_draw_value (GTK_SCALE (priv->time_scale), FALSE);
 	gtk_widget_set_size_request (priv->time_scale, EGG_PLAYER_PREVIEW_WIDTH, -1);
-	priv->time_label = gtk_label_new ("0:00");
-	gtk_misc_set_alignment (GTK_MISC (priv->time_label), 0.0, 0.5);
-
-	gtk_box_pack_start (GTK_BOX (hbox), align, FALSE, FALSE, 0);
-	gtk_box_pack_start (GTK_BOX (hbox), priv->time_scale, TRUE, TRUE, 0);
-	gtk_box_pack_start (GTK_BOX (hbox), priv->time_label, FALSE, FALSE, 0);
+	g_object_set (priv->time_scale,
+				  "valign", GTK_ALIGN_CENTER,
+				  "hexpand", TRUE,
+				  NULL);
+	gtk_grid_attach (grid, priv->time_scale, 2, 0, 1, 2);
+	priv->time_label = gtk_label_new ("0:00/0:00");
+	g_object_set (priv->time_label,
+				  "halign", GTK_ALIGN_END,
+				  "justify", GTK_JUSTIFY_RIGHT,
+				  "valign", GTK_ALIGN_CENTER,
+				  "xalign", 1.0, /* This is needed to right justify the label even though it is deprecated */
+				  NULL);
+	gtk_grid_attach (grid, priv->time_label, 3, 0, 1, 2);
 
 	g_signal_connect (G_OBJECT (priv->play_button), "clicked",
 					  G_CALLBACK (_clicked_cb), play_preview);
 	g_signal_connect (G_OBJECT (priv->time_scale), "change-value",
 					  G_CALLBACK (_change_value_cb), play_preview);
+	g_signal_connect (play_preview, "notify::duration",
+					  G_CALLBACK (_duration_notify_cb), NULL);
+	g_signal_connect (priv->time_label, "style-updated",
+					  G_CALLBACK (_style_updated_cb), play_preview);
 
-	gtk_box_pack_start (GTK_BOX (play_preview), vbox, FALSE, FALSE, 0);
-	gtk_box_pack_start (GTK_BOX (play_preview), hbox, FALSE, FALSE, 0);
+	gtk_box_pack_start (GTK_BOX (play_preview), GTK_WIDGET (grid), FALSE, FALSE, 0);
 
 	_ui_set_sensitive (play_preview, FALSE);
 
@@ -332,20 +354,10 @@ egg_play_preview_dispose (GObject *object)
 
 	_clear_pipeline (play_preview);
 
-	if (priv->title) {
-		g_free (priv->title);
-		priv->title = NULL;
-	}
-
-	if (priv->album) {
-		g_free (priv->album);
-		priv->album = NULL;
-	}
-
-	if (priv->uri) {
-		g_free (priv->uri);
-		priv->uri = NULL;
-	}
+	g_clear_pointer (&priv->title, g_free);
+	g_clear_pointer (&priv->artist, g_free);
+	g_clear_pointer (&priv->album, g_free);
+	g_clear_pointer (&priv->uri, g_free);
 
 	if (priv->timeout_id != 0) {
 		g_source_remove (priv->timeout_id);
@@ -427,6 +439,97 @@ egg_play_preview_get_property (GObject     *object,
 	}
 }
 
+static void
+calculate_widths (GtkWidget *label, int *widths)
+{
+	int i = 0;
+	for (i = 0; i < 10; i++) {
+		gchar *s;
+		s = g_strdup_printf ("%d", i);
+		gtk_label_set_text (GTK_LABEL (label), s);
+		gtk_widget_get_preferred_width (label, NULL, &widths[i]);
+		g_free (s);
+	}
+}
+
+static int
+get_widest_digit (int *widths, int max)
+{
+	int max_width = 0;
+	int i_max = 0;
+	int i;
+
+	for (i = 0; i <= max; i++) {
+		if (widths[i] > max_width) {
+			max_width = widths[i];
+			i_max = i;
+		}
+	}
+
+	return i_max;
+}
+
+static gchar *
+get_widest_time (int *widths, int duration)
+{
+	int minutes;
+	int seconds;
+	int m, s1, s2;
+	const char *format = "%d:%d%d/%d:%02d";
+
+	minutes = duration / 60;
+	seconds = duration % 60;
+
+	s1 = get_widest_digit (widths, 9);
+	if (minutes == 0)
+		s2 = get_widest_digit (widths, seconds % 10);
+	else
+		s2 = get_widest_digit (widths, 5);
+	if (minutes > 9) {
+		m = get_widest_digit (widths, 9);
+		m += get_widest_digit (widths, minutes % 10) * 10;
+		format = "%02d:%d%d/%d:%02d";
+	} else {
+		m = get_widest_digit (widths, minutes);
+	}
+
+	return g_strdup_printf (format, m, s2, s1, minutes, seconds);
+}
+
+static void
+set_time_label_width (EggPlayPreview *play_preview)
+{
+	EggPlayPreviewPrivate *priv;
+	int widths[10];
+	int w;
+	gchar *s;
+
+	priv = GET_PRIVATE (play_preview);
+	g_object_set (priv->time_label, "width-request", -1, NULL);
+	calculate_widths (priv->time_label, widths);
+	s = get_widest_time (widths, priv->duration / GST_SECOND);
+	gtk_label_set_text (GTK_LABEL (priv->time_label), s);
+	gtk_widget_get_preferred_width (priv->time_label, NULL, &w);
+	g_object_set (priv->time_label, "width-request", w, NULL);
+	g_free (s);
+	_ui_update_duration (play_preview);
+}
+
+static void
+_duration_notify_cb (EggPlayPreview *play_preview,
+					 GParamSpec     *pspec,
+					 gpointer        user_data)
+{
+	set_time_label_width (play_preview);
+}
+
+static void
+_style_updated_cb (GtkWidget *widget,
+				   gpointer   user_data)
+{
+	set_time_label_width (user_data);
+}
+
 static gboolean
 _timeout_cb (EggPlayPreview *play_preview)
 {
@@ -434,10 +537,19 @@ _timeout_cb (EggPlayPreview *play_preview)
 
 	priv = GET_PRIVATE (play_preview);
 
-	priv->position = _query_position (priv->playbin);
+	if (priv->seeking)
+		return TRUE;
+
+	if (gst_element_query (priv->playbin, priv->query)) {
+		gst_query_parse_position (priv->query, NULL, &priv->position);
+	} else {
+		return TRUE;
+	}
+
 	g_object_notify (G_OBJECT (play_preview), "position");
 
 	gtk_range_set_value (GTK_RANGE (priv->time_scale), priv->position * (100.0 / priv->duration));
+	_ui_update_duration (play_preview);
 
 	return TRUE;
 }
@@ -450,9 +562,9 @@ _ui_update_duration (EggPlayPreview *play_preview)
 
 	priv = GET_PRIVATE (play_preview);
 
-	str = g_strdup_printf ("%u:%02u",
-						   priv->duration / 60,
-						   priv->duration % 60);
+	str = g_strdup_printf ("%u:%02u/%u:%02u",
+						   (int) (priv->position / 60 / GST_SECOND), (int) (priv->position / GST_SECOND) % 60,
+						   (int) (priv->duration / 60 / GST_SECOND), (int) (priv->duration / GST_SECOND) % 60);
 
 	gtk_label_set_text (GTK_LABEL (priv->time_label), str);
 	g_free (str);
@@ -488,15 +600,36 @@ _ui_set_sensitive (EggPlayPreview *play_preview, gboolean sensitive)
 	gtk_widget_set_sensitive (priv->time_scale, sensitive && priv->is_seekable);
 }
 
+/*
+ * Note that handling ‘change-value’ is preferable to handling
+ * ‘value-changed’ as ‘change-value’ is only emitted in response to
+ * scroll events so it avoids a feedback loop which would exist
+ * between a ‘value-changed’ handler and timeout_cb() which calls
+ * gtk_range_set_value().
+ */
 static gboolean
 _change_value_cb (GtkRange *range, GtkScrollType scroll, gdouble value, EggPlayPreview *play_preview)
 {
 	EggPlayPreviewPrivate *priv;
+	GtkAdjustment *adjustment;
+	double lower, upper;
 
 	priv = GET_PRIVATE (play_preview);
+	adjustment = gtk_range_get_adjustment (range);
+	lower = gtk_adjustment_get_lower (adjustment);
+	upper = gtk_adjustment_get_upper (adjustment);
+	/* Clamp value to be within the adjustment range. */
+	if (value < lower)
+		value = lower;
 
-	if (priv->is_seekable)
-		_seek (priv->playbin, (gint) ((value / 100.0) * priv->duration));
+	if (value > upper)
+		value = upper;
+
+	if (priv->is_seekable && value != gtk_adjustment_get_value (adjustment)) {
+		priv->position = (gint64) (value / 100.0 * priv->duration);
+		_seek (play_preview, priv->playbin, priv->position);
+		_ui_update_duration (play_preview);
+	}
 
 	return FALSE;
 }
@@ -549,6 +682,7 @@ _setup_pipeline (EggPlayPreview *play_preview)
 	gst_object_unref (bus);
 
 	gst_element_set_state (priv->playbin, GST_STATE_NULL);
+	priv->query = gst_query_new_position (GST_FORMAT_TIME);
 }
 
 static void
@@ -567,6 +701,8 @@ _clear_pipeline (EggPlayPreview *play_preview)
 		gst_element_set_state (priv->playbin, GST_STATE_NULL);
                 gst_object_unref (GST_OBJECT (priv->playbin));
 		priv->playbin = NULL;
+		gst_query_unref (priv->query);
+		priv->query = NULL;
 	}
 }
 
@@ -574,8 +710,8 @@ static gboolean
 _process_bus_messages (GstBus *bus, GstMessage *msg, EggPlayPreview *play_preview)
 {
 	EggPlayPreviewPrivate *priv;
-	GstFormat format = GST_FORMAT_TIME;
-	GstTagList *tag_list;
+	GstTagList *tag_list = NULL;
+	gchar *title = NULL, *artist = NULL, *album = NULL;
 	gint64 duration;
 	GstState state;
 	GstStateChangeReturn result;
@@ -583,17 +719,15 @@ _process_bus_messages (GstBus *bus, GstMessage *msg, EggPlayPreview *play_previe
 	priv = GET_PRIVATE (play_preview);
 
 	switch (GST_MESSAGE_TYPE (msg)) {
-	case GST_MESSAGE_DURATION:
-		gst_message_parse_duration (msg, &format, (gint64*) &duration);
-
-		if (format != GST_FORMAT_TIME)
+	case GST_MESSAGE_DURATION_CHANGED:
+		if (!gst_element_query_duration (priv->playbin, GST_FORMAT_TIME, &duration))
 			break;
 
-		priv->duration = duration / GST_SECOND;
+		if (priv->duration == duration)
+			break;
 
+		priv->duration = duration;
 		g_object_notify (G_OBJECT (play_preview), "duration");
-
-		_ui_update_duration (play_preview);
 		break;
 
 	case GST_MESSAGE_EOS:
@@ -602,17 +736,42 @@ _process_bus_messages (GstBus *bus, GstMessage *msg, EggPlayPreview *play_previe
 
 	case GST_MESSAGE_TAG:
 		gst_message_parse_tag (msg, &tag_list);
-		gst_tag_list_get_string (tag_list, GST_TAG_TITLE, &priv->title);
-		gst_tag_list_get_string (tag_list, GST_TAG_ARTIST, &priv->artist);
-		gst_tag_list_get_string (tag_list, GST_TAG_ALBUM, &priv->album);
+		gst_tag_list_get_string (tag_list, GST_TAG_TITLE, &title);
+		gst_tag_list_get_string (tag_list, GST_TAG_ARTIST, &artist);
+		gst_tag_list_get_string (tag_list, GST_TAG_ALBUM, &album);
+		gst_tag_list_unref (tag_list);
 
-		g_object_notify (G_OBJECT (play_preview), "title");
-		g_object_notify (G_OBJECT (play_preview), "artist");
-		g_object_notify (G_OBJECT (play_preview), "album");
+		if (g_strcmp0 (title, priv->title) != 0) {
+			g_free (priv->title);
+			priv->title = title;
+			g_object_notify (G_OBJECT (play_preview), "title");
+		} else {
+			g_free (title);
+		}
+		if (g_strcmp0 (artist, priv->artist) != 0) {
+			g_free (priv->artist);
+			priv->artist = artist;
+			g_object_notify (G_OBJECT (play_preview), "artist");
+		} else {
+			g_free (artist);
+		}
+
+		if (g_strcmp0 (album, priv->album) != 0) {
+			g_free (priv->album);
+			priv->album = album;
+			g_object_notify (G_OBJECT (play_preview), "album");
+		} else {
+			g_free (album);
+		}
 
 		_ui_update_tags (play_preview);
 		break;
 
+	case GST_MESSAGE_ASYNC_DONE:
+		if (GST_OBJECT (priv->playbin) == GST_MESSAGE_SRC (msg))
+			priv->seeking = FALSE;
+		break;
+
 	case GST_MESSAGE_STATE_CHANGED:
 		result = gst_element_get_state (GST_ELEMENT (priv->playbin), &state, NULL, 500);
 
@@ -676,7 +835,7 @@ _query_seeking (GstElement *element)
 	return seekable;
 }
 
-static gint
+static gint64
 _query_duration (GstElement *element)
 {
 	GstStateChangeReturn result;
@@ -703,43 +862,18 @@ _query_duration (GstElement *element)
 
 	gst_element_set_state (element, state);
 
-	return (gint) (duration / GST_SECOND);
-}
-
-static gint
-_query_position (GstElement *element)
-{
-	gint64 position;
-
-	position = 0;
-
-	gst_element_query_position (element, GST_FORMAT_TIME, &position);
-
-	return (gint) (position / GST_SECOND);
+	return duration;
 }
 
 static void
-_seek (GstElement *element, gint position)
+_seek (EggPlayPreview *play_preview, GstElement *element, gint64 position)
 {
-	GstStateChangeReturn result;
-	GstState state;
+	EggPlayPreviewPrivate *priv;
 
-	result = gst_element_get_state (element, &state, NULL, GST_CLOCK_TIME_NONE);
-
-	if (result == GST_STATE_CHANGE_FAILURE)
-		return;
-
-	result = gst_element_set_state (element, GST_STATE_PAUSED);
-
-	if (result == GST_STATE_CHANGE_ASYNC) {
-		gst_element_get_state (element, NULL, NULL, GST_CLOCK_TIME_NONE);
-	}
-
-	gst_element_seek (element, 1.0, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH,
-					  GST_SEEK_TYPE_SET, position * GST_SECOND,
-					  GST_SEEK_TYPE_NONE, GST_CLOCK_TIME_NONE);
-
-	gst_element_set_state (element, state);
+	priv = GET_PRIVATE (play_preview);
+	if (gst_element_seek_simple (priv->playbin, GST_FORMAT_TIME,
+								 GST_SEEK_FLAG_FLUSH, position))
+		priv->seeking = TRUE;
 }
 
 static gboolean
@@ -849,7 +983,7 @@ egg_play_preview_set_uri (EggPlayPreview *play_preview, const gchar *uri)
 		_ui_set_sensitive (play_preview, TRUE);
 		_ui_update_duration (play_preview);
 		_ui_update_tags (play_preview);
-		priv->timeout_id = g_timeout_add_seconds (1, (GSourceFunc) _timeout_cb, play_preview);
+		priv->timeout_id = g_timeout_add (250, (GSourceFunc) _timeout_cb, play_preview);
 		g_source_set_name_by_id (priv->timeout_id, "[sound-juicer] _timeout_cb");
 	}
 
@@ -857,7 +991,7 @@ egg_play_preview_set_uri (EggPlayPreview *play_preview, const gchar *uri)
 }
 
 void
-egg_play_preview_set_position (EggPlayPreview *play_preview, gint position)
+egg_play_preview_set_position (EggPlayPreview *play_preview, gint64 position)
 {
 	EggPlayPreviewPrivate *priv;
 
@@ -867,7 +1001,7 @@ egg_play_preview_set_position (EggPlayPreview *play_preview, gint position)
 
 	/* FIXME: write function content */
 	if (priv->is_seekable) {
-		_seek (priv->playbin, MIN (position, priv->duration));
+		_seek (play_preview, priv->playbin, MIN (position, priv->duration));
 
 		g_object_notify (G_OBJECT (play_preview), "position");
 	}
@@ -921,7 +1055,7 @@ egg_play_preview_get_album (EggPlayPreview *play_preview)
 	return priv->album;
 }
 
-gint
+gint64
 egg_play_preview_get_position (EggPlayPreview *play_preview)
 {
 	EggPlayPreviewPrivate *priv;
@@ -933,7 +1067,7 @@ egg_play_preview_get_position (EggPlayPreview *play_preview)
 	return priv->position;
 }
 
-gint
+gint64
 egg_play_preview_get_duration (EggPlayPreview *play_preview)
 {
 	EggPlayPreviewPrivate *priv;
